#ifdef ALL
#define APPLICA
#define APPLICB
#define APPLICC
#define APPLICD
#define APPLICE
#endif
#ifndef APPLICA
#define LIBRARY_MODULE
#endif

#include <stdlib.h>
#include <time.h>
#include <string.h>
#ifdef __ZTC__
#include <disp.h>
#endif
#include "screen.hpp"
#include "applic.hpp"
#include "event.hpp"
#include "alert.hpp"
#include "help.hpp"


#ifdef APPLICA
ColorSet defaultColors = {
    Color(BrightWhite,Blue),    // Normal
    Color(Black,Blue),          // Disabled
    Color(Yellow,Cyan),         // Focus
    Color(Black,Cyan),          // Disabled + Focus
    Color(BrightWhite,Green),   // Highlight
    Color(Black,Green),         // Highlight + Disabled
    Color(Green,White),         // Highlight + Focus
    Color(Green,Black)          // Highlight + Disabled + Focus
    };
ColorSet BWdefaultColors = {
    Color(BrightWhite),         // Normal
    Color(White),               // Disabled
    Color(Gray,White),          // Focus
    Color(Gray,White),          // Disabled + Focus
    Color(Black,White),         // Highlight
    Color(Black,White),         // Highlight + Disabled
    Color(BrightWhite),         // Highlight + Focus
    Color(White)                // Highlight + Disabled + Focus
    };

ColorSet alertColors = {
    Color(BrightWhite,Red),     // Normal
    Color(Black,Red),           // Disabled
    Color(BrightWhite,Magenta), // Focus
    Color(Black,Magenta),       // Disabled + Focus
    Color(Red,White),           // Highlight
    Color(Red,Black),           // Highlight + Disabled
    Color(Magenta,White),       // Highlight + Focus
    Color(Magenta,Black)        // Highlight + Disabled + Focus
    };
ColorSet BWalertColors = {
    Color(BrightWhite),         // Normal
    Color(Black,White),         // Disabled
    Color(White),               // Focus
    Color(White),               // Disabled + Focus
    Color(Gray,White),          // Highlight
    Color(Black,White),         // Highlight + Disabled
    Color(Gray,White),          // Highlight + Focus
    Color(BrightWhite)          // Highlight + Disabled + Focus
    };

ColorSet helpColors = {
    Color(BrightWhite,Green),   // Normal
    Color(Black,Green),         // Disabled
    Color(BrightWhite,Cyan),    // Focus
    Color(Black,Cyan),          // Disabled + Focus
    Color(BrightWhite,Blue),    // Highlight
    Color(Black,Blue),          // Highlight + Disabled
    Color(Blue,White),          // Highlight + Focus
    Color(Blue,Black)           // Highlight + Disabled + Focus
    };

Application *application;
Screen *screen;
VObject *Application::kfocus, *Application::mfocus;
int Application::mrepeat;
Help *Application::helphandler;

Application::Application()
: window(0)
{
    ASSERT(application == 0);
    screen = new Screen();
    ASSERT(screen);
    application = this;
    if(!screen->iscolor())
        {                               // black and white
        defaultColors = BWdefaultColors;
        alertColors = BWalertColors;
        helpColors = BWalertColors;
        }
}

Application::~Application()
{
    ASSERT(application);
    if(Application::helphandler)
        {
        delete Application::helphandler;
        Application::helphandler = 0;
        }
    deleteall();
    application = 0;
    eventQueue.disable();
    delete screen;
}

implement(OCollection,Window)

#ifdef __TURBOC__
#include <time.h>
static void msleep(int nticks)
{
    nticks = (nticks * CLK_TCK) / 1000;
    clock_t start = clock();
    
    while(clock() - start < nticks)
        ;
}
#endif

void Application::enter(Window *control)
{
    eventQueue.enable();
    if(control)
        pop(control);
    while(size())
        {
        if(!kfocus)
        	{
	        Window *w = first();
	        if(!w->testcFlag(CFfocus))
	            w->setfocus();
			}

        Event *e = eventQueue();

        if(!e)
            {
            OIterator(Window) next(*this);
            Window *v;

            while(!e && (v = next()) != 0)
                {
                if(v->testcFlag(CFupdate))
                    v->update();
                e = eventQueue();
                }
            }
        if(mfocus && mrepeat && !e)
            {
            msleep(10);             // wait a little while for an event
            e = eventQueue();
            if(!e)
                {
                MouseMovement mouse;   // generate move event at last position

                mouse.despatch(control);
                }
            }
        if(e && e->despatch(control) == -1)
            return;
        }
}

void Application::paint(Rectangle r)
{
    r &= *screen;
    screen->fillrect(r,defaultColors.colors[SFnormal],' ');

    OIterator(Window) next(*this);
    Window *v;

    while((v = next()) != 0)
        if(v->intersects(r))
            {
            if(v->testcFlag(CFdirty))
                v->draw();
            else if(v->testcFlag(ControlFlag(CFpartdirty|CFcontentsdirty)))
                {
                v->dirty(r);
                v->update();
                }
            else
                v->paint(r);
            }
}

int Application::kbdevent(KeyboardEvent& e)
{
    if(Application::helphandler)
        switch(e.value())
            {
        case KF1:
            Application::helphandler->index();
            return 1;
        case KF9:
            if(!Application::help("Keys help"))
                Application::helphandler->index();
            return 1;
        case KShiftF10:
            if(!Application::help("Help for help"))
                Application::helphandler->index();
            return 1;
            }
    return 0;
}

int Application::dohotkey(KeyboardEvent& e)
{
    int r = 0;

    OIterator(Window) next(*this);
    Window *v;
    while((v = next()) != 0)
        {
        if((r = v->dohotkey(e)) != 0)
            return r;
        if(v == window)
            break;
        }
    return 0;
}

int Application::interruptevent(InterruptEvent &)
{
    static int inCtrlBreakHandler;

    if(!inCtrlBreakHandler)
        {
        inCtrlBreakHandler = 1;
        if(confirm("Ctrl-Break pressed\n\n"
                "Please confirm you\n"
                "wish to abandon\n"
                "the application"))
                    exit(0);
        inCtrlBreakHandler = 0;
        }
    return 1;
}

int Application::criterr(CriticalErrorEvent &e)
{
    static int inCritErrHandler;

    if(!inCritErrHandler)
        {
        inCritErrHandler = 1;
        alert(e.description());
        inCritErrHandler = 0;
        }
    return 1;
}

Window *Application::containingWindow(const Point &p)
{
    OIterator(Window) next(*this);
    Window *h;

    while((h = next()) != 0)
        {
        if(h->encloses(p))
            return h;
        if(h == window)
            break;
        }
    return 0;
}

void Application::pop(Window *w)
{
    addWindow(w);
    window = w;             // window has exclusive use of all events
}

void Application::addWindow(Window *w)
{
#ifndef NDEBUG
    if(VObject::debug)
        fprintf(VObject::debug,"Add %s to Application\n",w->inspectinfo());
#endif
    if(!screen->encloses(*w))
        w->placeInside(*screen);
    insert(w);
    w->setwindow(w);
    w->setfocus();
    w->dirty();
}

int Application::help(const char *key)
{
    int pageno;

    if(!Application::helphandler || (pageno = Application::helphandler->lookup(key)) < 0)
        return 0;
    return Application::helphandler->show(pageno);
}
#endif

#ifdef APPLICB
void Application::focusto(VObject *o)
{
    if(kfocus && kfocus != o)
        {
        kfocus->resetfocus();
        if(kfocus && kfocus->getwindow() && 
            (!o || kfocus->getwindow() != o->getwindow()))
                kfocus->getwindow()->resetfocus();
        }
#ifndef NDEBUG
    if(VObject::debug)
    	{
        fprintf(VObject::debug,"Focus changed from %s",kfocus ? kfocus->inspectinfo() : "NULL");
        fprintf(VObject::debug,"to %s\n",o ? o->inspectinfo() : "NULL");
        }
#endif
    kfocus = o;
}

void Application::mousefocusto(VObject *o,int repeat)
{
    if(mfocus && mfocus != o)
        mfocus->clearFlag(SFmouseflags);
    mfocus = o;
    mrepeat = repeat;
}

void Application::mousefocusoff(VObject *o)
{
    if(mfocus == o)
        {
        mfocus->clearFlag(SFmouseflags);
        mfocus = 0;
        mrepeat = 0;
        }
}
#endif

#ifdef APPLICC
int Application::openHelpFile(const char *filename, const char *progname)
{
    if(Application::helphandler)
        {
        delete Application::helphandler;
        }
    Application::helphandler = new Help(filename, progname);
    if(Application::helphandler->error())
        {
        delete Application::helphandler;
        Application::helphandler = 0;
        return 0;
        }
    return 1;
}
#endif

#ifdef APPLICD
int Application::removeWindow(Window *w)
{
    if(!OCollection(Window)::remove(w))
        return 0;
#ifndef NDEBUG
    if(VObject::debug)
        fprintf(VObject::debug,"Remove %s from Application\n",w->inspectinfo());
#endif
    if(window == w)
        window = 0;
    if(mfocus && mfocus->getwindow() == w)
        mfocus = 0;
    if(kfocus && kfocus->getwindow() == w)
        focusto(0);
    OIterator(Window) next(*this);
    Window *v;
    
    while((v = next()) != 0)
        v->dirty(*w);
    return 1;
}
#endif

#ifdef APPLICE
void Application::disable()
{
    eventQueue.disable();
    screen->disable();
}

void Application::enable()
{
    screen->enable();
    eventQueue.enable();
}
#endif
