#ifndef APPLIC_HPP
#define APPLIC_HPP

#ifndef WINDOW_HPP
#include "window.hpp"
#endif

declare(OCollection,Window);
declare(OIterator,Window);

/*
Every program that uses the toolkit must have one (and only one)
Application declared in it. In most programs, the Application is
declared in the main module immediately after the EventQueue, and
that is all there is to it. If you want to change the default
behaviour of the application, you may subclass Appliation, and
override some virtual methods in it. 
*/
class Application : OCollection(Window) {
friend class Window;
static VObject *kfocus, *mfocus;
static int mrepeat;
static class Help *helphandler;
	Window *window;
public:
	Application();
	~Application();

	Window *containingWindow(const Point &p);
	virtual void enter(Window *control = 0);
	void pop(Window *w);		// Window gets ALL input events
	void addWindow(Window *w);
	int removeWindow(Window *w);

	virtual void paint(Rectangle r);
	virtual int kbdevent(KeyboardEvent& e);
	virtual int dohotkey(KeyboardEvent& e);
	virtual int interruptevent(InterruptEvent& e);
	virtual int criterr(CriticalErrorEvent& e);

//~ Static methods
// A number of static methods are provided, for use either in
// VObject's or in your code. 
static void focusto(VObject *o);
/*
Move the keyboard focus to a VObject. Keyboard events will be passed
to the specified VObject until another VObject takes the focus. 
*/
static void mousefocusto(VObject *o,int repeat = 0);
/*
Move the mouse focus to a VObject. All mouse events will be passed
to the VObject until it calls mousefocusoff. 
*/
static void mousefocusoff(VObject *o);
/*
Mouse events will be passed to the topmost VObject containing the
mouse cursor. 
*/
static VObject *kbdfocus() { return kfocus; }
// The current value of the keyboard focus.
static VObject *moufocus() { return mfocus; }
// The current value of the mouse focus.
static int openHelpFile(const char *filename, const char *progname = 0);
/*
Open the help file filename.HLP, and direct future help requests
into this file. The table of contents and index at the beginning of
the file are read into memory. Returns 0 if the file was not found,
or was not a valid help file compiled with MAKEHELP.EXE. If progname
is supplied, it should be the value of argv[0] (i.e. the program
name), and the helpfile will be searched for on the program
directory. In any case, if the help file is not found, all
directories on the PATH are searched. 
*/
static int help(const char *key);
/*
The string key is looked up in the table of contents of the current
help file, and an appropriate page of help is displayed. On exit
from help, the function returns 1. If the key is not found in the
table of contents, no action is taken, and the function returns 0. 
*/
static void disable();		// Disable all devices ready for spawn.
static void enable();		// Re-enable all devices after spawn.
	};

extern Application *application;
// A pointer to the single Application object you declare in your main
// program, used by the toolkit internals.

#endif
