#ifdef ALL
#define COLLECTA
#define COLLECTB
#define COLLECTC
#define COLLECTD
#define COLLECTE
#define COLLECTF
#define COLLECTG
#define COLLECTH
#define COLLECTI
#define COLLECTJ
#define COLLECTK
#define COLLECTL
#define COLLECTM
#define COLLECTN
#define COLLECTO
#define COLLECTP
#define COLLECTQ
#define COLLECTR
#define COLLECTS
#define COLLECTT
#define COLLECTU
#define COLLECTV
#define COLLECTW
#endif
#ifndef COLLECTA
#define LIBRARY_MODULE
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "collect.hpp"
#include "vobject.hpp"

#ifdef COLLECTA
void *debugPointer;
int x;

int Collection::expand(int n)
{
    const int MAXGRAIN = 5;

    if((uelements + n) > nelements)
        {
        if(n == 1)
        	n = MAXGRAIN;
        
        int extras = uelements + n - nelements;

        int nbytes = (nelements + extras) * sizeof(void *);

        void **l;

        if(list == 0)
            l = (void **)malloc(nbytes);
        else
            l = (void **)realloc(list,nbytes);

        if(!l)
            {
            seterror();
            return 0;
            }
        if(l == debugPointer)
            x = n;
        list = l;
        nelements += extras;
        }
    return 1;
}
#endif

#ifdef COLLECTB
void Collection::appendvalist(va_list v)
{
    void *vp;

    while(!err && (vp = va_arg(v,void *)) != 0)
        append(vp);
}
#endif

#ifdef COLLECTC
Collection::Collection(int hintSize)
: list(0), nelements(0), uelements(0), err(0)
{
	if(hintSize > 0)
		expand(hintSize);
}
#endif

#ifdef COLLECTD
Collection::Collection(const Collection& c)
: list(0), nelements(0), uelements(0), err(0)
{
    *this = c;              // use operator=
}
#endif

#ifdef COLLECTE
Collection::~Collection()
{
    if(list)
        free(list);
}

#ifdef NDEBUG
void Collection::print(const char *s,const char newline) const
{
}
#else
void Collection::print(const char *s,const char newline) const
{
    int i;

    printf("%sCollection @%p%s[%d,%d]",s,this,err ? "ERR" : "",
        uelements,nelements);

    if(uelements)
        printf("%p",list[0]);
    for(i = 1; i < uelements; i++)
        printf(",%p",list[i]);
    putchar(newline);
}

#endif
int Collection::debug;
#endif

#ifdef COLLECTF
Collection& Collection::operator= (const Collection& c)
{
    if(this == &c)
        return *this;
    nelements = c.nelements;
    uelements = c.uelements;
    if(list)
        free(list);
    if(c.list == 0)
        list = 0;
    else
        {
        int nbytes = nelements * sizeof(void *);
        list = (void **)malloc(nbytes);
        if(list)
            memcpy(list,c.list,nbytes);
        else
            seterror();
        }
    return *this;
}
#endif

#ifdef COLLECTG
int Collection::add(void *c,int position)
{
    if(!expand())
        return -1;
    if(position < 0 || position > uelements)
        position = uelements;
    if(position < uelements)
        memmove(list + position + 1,list + position,(uelements - position) * sizeof(void *));
    list[position] = c;
    uelements++;
    return position;
}
#endif

#ifdef COLLECTH
void Collection::insert(void *c)
{
    if(!expand())
        return;
    if(uelements)
        memmove(list + 1,list,uelements * sizeof(void *));
    list[0] = c;
    uelements++;
}
#endif

#ifdef COLLECTI
void Collection::append(void *c)
{
    if(!expand())
        return;
    list[uelements++] = c;
}

Collection& Collection::operator+ (const Collection& c)
{
	expand(c.uelements);
    if(list)
    	{
        memcpy(list + uelements,c.list,c.uelements * sizeof(void *));
        uelements += c.uelements;
        }
    return *this;
}
#endif

#ifdef COLLECTJ
int Collection::remove(const void *c)
{
    int i = elementNo(c);

    if(i >= 0)
        {
        if(i < uelements - 1)
            memmove(list + i,list + i + 1,(uelements - i - 1) * sizeof(void *));
        uelements--;
        }
    return (i >= 0);
}
#endif

#ifdef COLLECTK
int Collection::elementNo(const void *e) const
{
    for(int i = 0; i < uelements; i++)
        if(list[i] == e)
            return i;
    return -1;
}
#endif

#ifdef COLLECTL
void *Collection::get()
{
    if(uelements)
        {
        void *v = list[0];
        memmove(list,list + 1,(uelements - 1) * sizeof(void *));
        uelements--;
        return v;
        }
    return 0;
}
#endif

#ifdef COLLECTM
static ObjectCollection *sortedCollection;

static int cdecl scompare(const void *v1,const void *v2)
{
    return sortedCollection->compare((VObject *)*((void **)v1),(VObject *)*((void **)v2));
}

void ObjectCollection::resort()
{
    if(size() > 0)
        {
        ASSERT(sortedCollection == 0 /* Recursive call to Collection::sort() */);
        sortedCollection = this;
        qsort(list,size(),sizeof(void**),scompare);
        sortedCollection = 0;
        }
}

int ObjectCollection::compare(const VObject *v1, const VObject *v2) const
{
    return compfn ? compfn(v1,v2) : 0;
}

int ObjectCollection::add(VObject *c)
{
    if(!expand())
        return -1;
    for(int i = 0; i < size() && compare(c,(VObject *)list[i]) >= 0; i++)
        ;                       // find where to put it
    Collection::add(c,i);
    return i;
}
#endif

#ifdef COLLECTN
Iterator::Iterator(const Collection& l,int backwards)
: list(l), i(-1), b(backwards)
{
}
#endif

#ifdef COLLECTO
Iterator::Iterator(const Collection& l,const void *e,int backwards)
: list(l), b(backwards)
{
    set(e);
}
#endif

#ifdef COLLECTP
int Iterator::set(const void* s)
{
    i = list.elementNo(s);
    if(i == -1)
        {
        if(b)
            i = list.size();
        return 0;
        }
    return 1;
}
#endif

#ifdef COLLECTQ
void *Iterator::next()
{
    void *v;

    if(list.size())
        {
        if(i == list.size() - 1)
            {
            i = -1;
            v = 0;
            }
        else
            v = list.list[++i];
        }
    else
        v = 0;
    return v;
}
#endif

#ifdef COLLECTR
void *Iterator::previous()
{
    void *v;

    if(list.size())
        {
        if(i == -1)
            i = list.size();
        if(i--)
            v = list.list[i];
        else
            v = 0;
        }
    else
        v = 0;
    return v;
}

void *Iterator::operator()()
{
    return (b) ? previous() : next();
}
#endif

#ifdef COLLECTS
ObjectCollection::ObjectCollection(int hintSize)
: Collection(hintSize), compfn(0)
{
}
#endif

#ifdef COLLECTT
ObjectCollection::ObjectCollection(const ObjectCollection& c)
: Collection(c), compfn(c.compfn)
{
}
#endif

#ifdef COLLECTU
void ObjectCollection::deleteall()
{
    VObject *v;

    while((v = getlast()) != 0)
		{
		VALIDATE(v);
        delete v;
		}
}
#endif

#ifdef COLLECTV
ObjectIterator::ObjectIterator(const ObjectCollection& s,int bk)
: Iterator(s,bk)
{
}
#endif

#ifdef COLLECTW
ObjectIterator::ObjectIterator(const ObjectCollection& s,const VObject *e,int bk)
: Iterator(s,e,bk)
{
}
#endif
