#ifdef ALL
#define GROUPA
#define GROUPB
#define GROUPC
#define GROUPD
#define GROUPE
#define GROUPF
#endif
#ifndef GROUPA
#define LIBRARY_MODULE
#endif

#include <stdarg.h>
#include "group.hpp"
#include "window.hpp"
#include "event.hpp"

#ifdef GROUPA
void Group::addvalist(VObject *v,va_list args)
{
    add1(v);
    while((v = va_arg(args,VObject *)) != EndList)
        add1(v);
    addDone();
}

void Group::add(const ObjectCollection &c)
{
	ObjectIterator next(c);
	VObject *v;

	objects.expand(c.size());		// optimisation
    while((v = next()) != 0)
        add1(v);
    addDone();
}
#endif

#ifdef GROUPB
Group::Group(const Rectangle &r,const char *n)
: VObject(r,n), kbdfocus(0)
{
}
#endif

#ifdef GROUPC
Group::Group(const char *n)
: VObject(n), kbdfocus(0)
{
}
#endif

#ifdef GROUPD
Group::Group(const coord width,const coord height,const char *n)
: VObject(width,height,n), kbdfocus(0)
{
}
#endif

#ifdef GROUPE
Group::Group(const coord l,const coord t,const coord r,const coord b,const char *n)
: VObject(l,t,r,b,n), kbdfocus(0)
{
}
#endif

#ifdef GROUPF
Group::~Group()
{
    objects.deleteall();
}

void Group::contentsSet(VObject* o,ControlFlag f)
{
    if(f & CFfocus)
        {
        VObject *k = kbdfocus;

        kbdfocus = o;
        if(testcFlag(CFfocus) && k && k != o)
            k->resetfocus();
        }
    VObject::contentsSet(o,f);
}

void Group::contentsDeleted(VObject* v)
{
    remove(v);
}

void Group::setwindow(Window *w)
{
    if(window == w)
        return;
    VObject::setwindow(w);
	ObjectIterator next(objects);
    VObject *v;

    while((v = next()) != 0)
        v->setwindow(w);
}

void Group::move(coord x,coord y)
{
    if(!x && !y)
        return;
    VObject::move(x,y);
	ObjectIterator next(objects);
    VObject *v;

    while((v = next()) != 0)
        v->move(x,y);
}

// Add & remove VObjects
void Group::add1(VObject *v)
{
    objects.append(v);
    v->setcontainer(this);
    v->setwindow(window);
    v->dirty();
    if(!testcFlag(CFactive) && v->testcFlag(CFactive))
        setcFlag(CFactive);
}

int Group::remove1(VObject *v)
{
    if(!v || !objects.remove(v))
        return 0;
    if(v->testcFlag(CFfocus))
        v->resetfocus();
    if(kbdfocus == v)
        clearfocus(v);
    v->setcontainer(0);
    v->setwindow(0);
    dirty(*v);
    return 1;
}

void Group::addDone()
{
}

void Group::removeDone()
{
	addDone();
}

void Group::update()
{
    VObject::update();
	ObjectIterator next(objects);
    VObject *v;

    while((v = next()) != 0)
        if(v->testcFlag(CFupdate))
            v->update();
    clean();
}

coord Group::minwidth() const
{
    ObjectIterator next(objects);
    VObject *v;

    for(coord w = 0; (v = next()) != 0; )
        {
        coord x = v->minwidth();
        if(w < x)
            w = x;
        }
    return w;
}

coord Group::minheight() const
{
    ObjectIterator next(objects);
    VObject *v;

    for(coord h = 0; (v = next()) != 0; )
        {
        coord y = v->minheight();
        if(h < y)
            h = y;
        }
    return h;
}

coord Group::maxwidth() const
{
    ObjectIterator next(objects);
    VObject *v;

    for(coord w = 0; (v = next()) != 0; )
        {
        coord x = v->maxwidth();
        if(w < x)
            w = x;
        }
    return w;
}

coord Group::maxheight() const
{
    ObjectIterator next(objects);
    VObject *v;

    for(coord h = 0; (v = next()) != 0; )
        {
        coord y = v->maxheight();
        if(h < y)
            h = y;
        }
    return h;
}

void Group::paint(Rectangle r)
{
    r &= *this;
    VObject::paint(r);
    paintcontents(r);
}

void Group::paintcontents(Rectangle &r)
{
	ObjectIterator next(objects);
    VObject *v;

    while((v = next()) != 0)
        if(v->intersects(r))
            {
            if(v->testcFlag(CFdirty))
                v->draw();
			else if(v->testcFlag(ControlFlag(CFpartdirty|CFcontentsdirty)))
                {
                v->dirty(r);
                v->update();
                }
            else
                v->paint(r);
            }
}

int Group::setfocus(int tabstop)
{
    if(!tabstop && testFlag(SFtabstop))
        return 0;
#ifndef NDEBUG
    if(debug)
        fprintf(debug,"%s->setfocus\n",debuginfo());
#endif
    if(kbdfocus && kbdfocus->setfocus(tabstop))
        return 1;
	ObjectIterator next(objects);
    VObject *v;
    
    while((v = next()) != 0)
        if(v->setfocus(tabstop))
            return 1;
    return 0;
}

void Group::resetfocus()
{
    if(kbdfocus && kbdfocus->testcFlag(CFfocus))
        kbdfocus->resetfocus();
    VObject::resetfocus();
    clearFlag(SFselected);
}

void Group::clearfocus(VObject *v)
{
    if(kbdfocus == v)
        kbdfocus = v = 0;
    if(!v)
        VObject::clearfocus();
}

int Group::incfocus(int tabstop,int decrement)
{
    switch(objects.size())
        {
    case 0:
        return 0;
    case 1:
        return objects.first()->incfocus(tabstop,decrement);
        }
#ifndef NDEBUG
    if(debug)
        fprintf(debug,"%s->incfocus(%sTabstop,%srement)\n",debuginfo(),
            tabstop ? "" : "No",decrement ? "dec" : "inc");
#endif
	ObjectIterator next(objects,kbdfocus,decrement);
    VObject *v;

    for(v = next(); v != kbdfocus; v = next())
        {
        if(!v)
            {
            next.reset();
            continue;       // wrap round
            }
        if(v->setfocus(tabstop))
            return 1;
        }
    if(kbdfocus)
        return kbdfocus->incfocus(tabstop,decrement);
    return 0;
}

int Group::nextTabstop(int decrement,int wrap)
{
    if(testFlag(SFtabstop))
        return VObject::nextTabstop(decrement,wrap);
    switch(objects.size())
        {
    case 0:
        return 0;
    case 1:
        return objects.first()->nextTabstop(decrement,wrap);
        }

    VObject *start = testcFlag(CFfocus) ? kbdfocus : 0;

    if(start && start->nextTabstop(decrement))  // try focussed item first
        return 1;

	ObjectIterator next(objects,start,decrement);   // try later items
    VObject *v;

    if(!wrap)
        start = 0;
    for(v = next(); v != start; v = next())
        {
        if(!v)
            {
            next.reset();
            continue;
            }
        if(v->nextTabstop(decrement))
            return 1;
        }
    return 0;
}

VObject *Group::containingObject(const Point &p)
{
	ObjectIterator next(objects);
    VObject *h;

    while((h = next()) != 0)
        if(h->encloses(p))
            return h->containingObject(p);
    return VObject::containingObject(p);
}

int Group::dohotkey(KeyboardEvent& e)
{
    int r = 0;

	ObjectIterator next(objects);
    VObject *v;
    while((v = next()) != 0)
        if((r = v->dohotkey(e)) != 0)
            return r;
    return VObject::dohotkey(e);
}

#ifdef NINSPECT
char *Group::inspectinfo(char *buf,const char *s) const
{
    return "";
}
#else
char *Group::inspectinfo(char *buf,const char *s) const
{
    if(!s)
        s = "Group ";
    buf = VObject::inspectinfo(buf,s);
	ObjectIterator next(objects);
    VObject *v;

    strcat(buf,"[");
    if((v = next()) != 0)
        {
        strcat(buf,form("%s@%p",v->name(),v));
        while((v = next()) != 0)
            strcat(buf,form(",%s@%p",v->name(),v));
        }
    return strcat(buf,"]\n");
}
#endif
#endif
