static char RCSid[] = "$RCSfile: help.cpp $ $Revision: 1.6 $";
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include "help.hpp"
#include "dialog.hpp"
#include "event.hpp"
#include "applic.hpp"
#include "flist.hpp"

static Help *helphandler;
const int maxLen = 256;
static char inbuf[maxLen+2];

implement(LCollection,HelpKey)

static int getaline(FILE *f)
{
    if(!fgets(inbuf,maxLen,f))
        return 0;

    char *e = strchr(inbuf,'\n');

    if(e)
        *e = 0;
    return 1;
}

static const char *normalise(const char *s)
{
    while(isspace(*s))
        s++;

    static char string[80];
    char *d = strchr(s,'~');

    strcpy(string,s);
    if(d)
        strcpy(string + (d - (char *)s),d + 1);
    d = string + strlen(string);
    while(--d > string && (*d == ' ' || *d == '.'))
        *d = 0;
    return string;
}

class HelpXref : public Button
    {
friend class HelpPage;
    HelpXref(int x, int y, int len, int pageno, const char *line);

    virtual int setpressed();
    virtual coord minwidth() const;
    virtual void paint(Rectangle r);
    virtual int kbdevent(KeyboardEvent& e);

    int pageno, len;
    };

class HelpPage : public Group
    {
friend class Help;
    HelpPage();
    ~HelpPage();

    virtual coord minwidth() const;
    virtual coord minheight() const;
    virtual coord maxwidth() const;
    virtual coord maxheight() const;
    virtual void paint(Rectangle r);
    
    void deleteall();
    int readpage(FILE *file,long offset);
    
    Collection lines;
    int w, h;
    };

class HelpWindow : public Dialog {
public:
    HelpWindow(VObject *v,int pageno = -1);

    virtual int setfocus(int tabstop = 1);
    virtual int kbdevent(KeyboardEvent &e);
    virtual int leftclick(MouseEvent &e);
    virtual int rightclick(MouseEvent &e);

private:
    int pageno;
    };

class HelpIndex : public List {
public:
    HelpIndex(LCollection(HelpKey) &keys);
    ~HelpIndex();

    virtual int pick();             // an item was selected with <cr> or click
    };

static int sortkeys(const HelpKey *k1,const HelpKey *k2)
{
#ifdef __ZTC__
    return strcmpl(k1->name(),k2->name());
#else
    return stricmp(k1->name(),k2->name());
#endif
}

Help::Help(const char *helpfile, const char *progname)
: pages(0), npages(0), cpage(-1)
{
    int nkeys;

    file = fopen(form("%s.HLP",helpfile),"rb");
	if(progname && !file)
		file = fopen(form("%s.HLP",fullfilename(dir(progname),helpfile)),"rb");
	if(!file)
		{
		char *p = getenv("PATH");

		if(p && *p)
			{
			static char path[512];

			strcpy(path,p);
			p = path;
			while(*p)
				{
				char *q = p;
				
				p = strchr(p,';');
				if(!p)
					p = "";
				else
					*p++ = 0;
				file = fopen(form("%s.HLP",fullfilename(q,helpfile)),"rb");
				if(file)
					break;
				}
			}
		}
    if(file && !readheader())
        {
        fclose(file);
        if(pages)
            delete pages;
        file = 0;
        pages = 0;
        npages = 0;
        }
    if(file)
        {
        keys.sort(sortkeys);
        helphandler = this;
        }
}

Help::~Help()
{
    if(file)
        fclose(file);
	if(pages)
    	delete pages;
    keys.deleteall();
    helphandler = 0;
}

int Help::lookup(const char *key)
{
    LIterator(HelpKey) next(keys);
    HelpKey *ky;

    key = normalise(key);
    while((ky = next()) != 0)
        if(*ky == key)
            return ky->page();
    return -1;
}

int Help::readheader()
{
    int nkeys;

    if(fscanf(file,"%d %d",&npages,&nkeys) != 2)
        return 0;
    pages = new long[npages+1];
    if(!pages)
        return 0;
    // read page offsets
    for(int i = 0; i < npages; i++)
        if(fscanf(file,"%6ld",&pages[i]) != 1)
            return 0;
    // read key table
    for(i = 0; i < nkeys; i++)
        {
        int page;
        char c;

        if(fscanf(file,"%d%c",&page,&c) != 2 || c != ' ')
            return 0;
        if(!getaline(file))
            return 0;
        keys.append(new HelpKey(inbuf,page));
        }
    return 1;
}

int Help::show(int pageno)
{
    if(pageno < 0 || pageno >= npages)
        return 0;

    HelpPage *p = new HelpPage;

	if(!p)
		return 0;
    if(!p->readpage(file,pages[pageno]))
        {
        delete p;
        return 1;
        }

    HelpWindow *w = new HelpWindow(p,pageno);

    w->enter();
    return 1;
}

void Help::index()
{
    HelpWindow *w = new HelpWindow(new HelpIndex(keys));

    w->enter();
}

HelpWindow::HelpWindow(VObject *v,int pageno)
: Dialog(v,v->name()), pageno(pageno)
{
    setcolors(helpColors);
    recogniseEvents();
}

int HelpWindow::setfocus(int tabstop)
{
    if(Dialog::setfocus(tabstop))
        return 1;
    return VObject::setfocus(tabstop);
}

int HelpWindow::kbdevent(KeyboardEvent &e)
{
    switch(e.value())
        {
    case Kesc:
        closeok();
        break;
    case KF9:
        if(Application::help("Keys help"))
            window->closeok();
        break;
    case KF1:
        if(pageno >= 0)
            {
            window->closeok();
            helphandler->index();
            break;
            }
        /* drop through */
    case KShiftF10:
        if(Application::help("Help for help"))
            window->closeok();
        break;
    case Kpgup:
        if(pageno > 0)
            {
            window->closeok();
            helphandler->show(pageno-1);
            }
        break;
    case Kpgdn:
        if(pageno < helphandler->maxpage())
            {
            window->closeok();
            helphandler->show(pageno+1);
            }
        break;
    default:
        return Dialog::kbdevent(e);
        }
    return 1;
}

int HelpWindow::leftclick(MouseEvent &e)
{
    if(pageno >= 0 && e.at().y == top() && this->encloses(e.at()))
        {
        window->closeok();
        helphandler->index();
        return 1;
        }
    return Dialog::leftclick(e);
}

int HelpWindow::rightclick(MouseEvent &)
{
    closeok();
    return 1;
}

HelpKey::HelpKey(const char *key,int pageno)
: ListElement(key), p(pageno)
{
	k = new char[strlen(key)+1];
    if(k)
		{
        strcpy(k,key);
		newname(k);
		}
}

HelpKey::~HelpKey()
{
	newname("HelpKey");
    delete k;
}

HelpXref::HelpXref(int x, int y, int len, int pageno, const char *line)
: Button(len,1,line+x), pageno(pageno), len(len)
{
    position(x,y);
    setFlag(SFtabstop);
}

int HelpXref::setpressed()
{
    window->closeok();
    helphandler->show(pageno);
    return 1;
}

int HelpXref::kbdevent(KeyboardEvent& e)
{
    int r;

    switch(e.value())
        {
        case '\n':
        case '\r':
            if(!testFlag(SFdisabled))
                {
                setpressed();
                r = 1;
                }
            break;
        case KF1:
            r = 0;
            break;
        default:
            r = Control::kbdevent(e);
        }
    return r;
}

coord HelpXref::minwidth() const
{
    return len;
}

void HelpXref::paint(Rectangle r)
{
    Button::paint(r);
    window->printstring(topleft(),r,getcolor(),name(),len);
}

HelpPage::HelpPage()
: Group("Help"), w(0), h(0)
{
}

HelpPage::~HelpPage()
{
    deleteall();
}

coord HelpPage::minwidth() const
{
    return w;
}

coord HelpPage::minheight() const
{
    return h;
}

coord HelpPage::maxwidth() const
{
    return w;
}

coord HelpPage::maxheight() const
{
    return h;
}

void HelpPage::paint(Rectangle r)
{
    int col = getcolor() | SFdisabled;

    r &= *this;
    window->fillrect(r,col,' ');

    Iterator next(lines);
    void *v;

    for(Point p = topleft(); (v = next()) != 0; p.y++)
        window->printstring(p,r,col,(const char *)v);
    Group::paintcontents(r);
}

void HelpPage::deleteall()
{
    void *v;

    while((v = lines.getlast()) != 0)
        delete (char *)v;
}

int HelpPage::readpage(FILE *file,long offset)
{
    int xref;
    char c;

    deleteall();            // delete lines
    objects.deleteall();    // delete xrefs
    fseek(file,offset,0);
    if(fscanf(file,"%d %d %d%c",&w,&h,&xref,&c) != 4 || c != '\n')
        return 0;
    setsize(w,h);
    for(int i = 0; i < h; i++)
        {
        if(!getaline(file))
            return 0;
        char *l = new char[strlen(inbuf)+1];
        if(!l)
            return 0;
        strcpy(l,inbuf);
        lines.append(l);
        }
    for(i = 0; i < xref; i++)
        {
        int len, x, y, pageno;
        
        if(fscanf(file,"%d %d %d %d",&len,&x,&y,&pageno) != 4)
            return 0;
        add(new HelpXref(x,y,len,pageno,(char *)lines[y]));
        }
    return 1;
}

HelpIndex::HelpIndex(LCollection(HelpKey) &keys)
: List(keys,"Help Index")
{
}

HelpIndex::~HelpIndex()
{
    deselect();
}

int HelpIndex::pick()
{
    int l = line();

    if(l < 0)
        return 0;
    window->closeok();
    helphandler->show(((HelpKey *)(collection[l]))->page());
    return 1;
}

int showHelpMenu(const char *o)
{
    if(helphandler)
        {
        int pageno = helphandler->lookup(o);
        
        if(pageno >= 0)
            helphandler->show(pageno);
        else
            helphandler->index();
        return 1;
        }
    return 0;
}
