// Help system header
#ifndef HELP_HPP
#define HELP_HPP
#ifndef GROUP_HPP
#include "group.hpp"
#endif
#ifndef BUTTON_HPP
#include "button.hpp"
#endif
#ifndef LIST_HPP
#include "list.hpp"
#endif
#ifndef BUFSIZ
#include <stdio.h>
#endif

extern ColorSet helpColors;

/*
Part of the help system, this class holds details of a key in the
help index. 
*/
class HelpKey : public ListElement {
friend class Help;
public:
	int compare(const char *key) const { 
#ifdef __ZTC__
	return strcmpl(k,key); 
#endif
#ifdef __TURBOC__
	return strcmpi(k,key); 
#endif
	}
	int operator==(const char *key) const { return !compare(key); }
	int page() const { return p; }
	const char *key() const { return k; }

private:
	HelpKey(const char *key,int pageno);
	~HelpKey();

	char *k;
	int p;
	};

declare(LCollection,HelpKey);
declare(LIterator,HelpKey);

/*
This class (together with the related classes whose names begin with
"Help") provides the functionality of the context-sensitive, hyper-
text help system. It contains the help index for a particular help
file, and has methods to look up help pages etc. Help files are
created using any text editor, and "compiled" using MAKEHELP.EXE. 
*/
class Help {
public:
	Help(const char *helpfile, const char *progname = 0);
	~Help();

	int error() const { return !file; }
	int maxpage() const { return npages - 1; }
	int lookup(const char *key);	// -1 if not found
	int readheader();
	int show(int pageno);
	void index();

private:
	LCollection(HelpKey) keys;
	long *pages;
	FILE *file;
	int npages, cpage;
	};

#endif
