#ifdef ALL
#define RECTA
#define RECTB
#define RECTC
#define RECTD
#define RECTE
#define RECTF
#define RECTG
#define RECTH
#define RECTI
#define RECTJ
#define RECTK
#define RECTL
#define RECTM
#define RECTN
#endif
#ifndef RECTA
#define LIBRARY_MODULE
#endif

/*
 * Basic point and rectangle classes
 */
#include <stdio.h>
#include "rect.hpp"


#ifdef RECTA
#ifdef NINSPECT
char *Point::inspectinfo(char *buf,const char *s) const
{
	return "";
}

char *Rectangle::inspectinfo(char * buf,const char *s) const
{
	return "";
}
#else
static char inspectbuf[512];

char *Point::inspectinfo(char *buf,const char *s) const
{
	if(!s)
		s = "Point";
	if(!buf)
		{
		inspectbuf[0] = 0;
		buf = inspectbuf;
		}
	return strcat(buf,form("%s(%d,%d)",s,x,y));
}

char *Rectangle::inspectinfo(char * buf,const char *s) const
{
	if(!s)
		s = "Rectangle";
	if(!buf)
		{
		inspectbuf[0] = 0;
		buf = inspectbuf;
		}
	return strcat(buf,form("%s(%d,%d,%d,%d)",s,left(),top(),right(),bottom()));
}
#endif

void Rectangle::build(const coord left,const coord top,const coord right,const coord bottom)
{
	ASSERT(left <= right && top <= bottom);
	tl.x = left; tl.y = top; 
	br.x = right; br.y = bottom;
}
#endif

#ifdef RECTB
Rectangle::Rectangle()
{
	tl.x = tl.y = br.x = br.y = 0;
}
#endif

#ifdef RECTC
Rectangle::Rectangle(const coord l,const coord t,const coord r,const coord b)
{
	build(l,t,r,b);
}
#endif

#ifdef RECTD
Rectangle::Rectangle(const Point& topl,const Point& botr)
{
	build(topl.x,topl.y,botr.x,botr.y);
}
#endif

#ifdef RECTE
Rectangle::Rectangle(const Point& topl,const coord width,const coord height)
{
	build(topl.x,topl.y,topl.x+width-1,topl.y+height-1);
}
#endif

#ifdef RECTF
Rectangle::Rectangle(const coord width,const coord height)
{
	build(0,0,width > 0 ? width - 1 : 0,height > 0 ? height - 1 : 0);
}
#endif

#ifdef RECTG
Rectangle::Rectangle(const Rectangle &r)
{
	build(r.tl.x,r.tl.y,r.br.x,r.br.y);
}
#endif

#ifdef RECTH
// Changing the Rectangle
void Rectangle::position(coord x,coord y)
{
	br.x += x - tl.x;
	br.y += y - tl.y;
	tl.x = x;
	tl.y = y;
}
#endif

#ifdef RECTI
void Rectangle::move(coord x,coord y)
{
	tl.x += x;
	tl.y += y;
	br.x += x;
	br.y += y;
}
#endif

#ifdef RECTJ
void Rectangle::setwidth(const coord w)
{
	ASSERT(w > 0);
	br.x = tl.x + w - 1;
}

void Rectangle::setheight(const coord h)
{
	ASSERT(h > 0);
	br.y = tl.y + h - 1;
}
#endif

#ifdef RECTK
// Boolean methods
int Rectangle::encloses(const coord x,const coord y) const
{
	return (x >= left() && x <= right() && y >= top() && y <= bottom());
}
#endif

#ifdef RECTL
int Rectangle::encloses(const Point& p) const		// is p in this
{
	return (p >= tl && p <= br);
}

int Rectangle::encloses(const Rectangle& r) const	// is r in this
{
	return (r.tl >= tl && r.br <= br);
}

int Rectangle::intersects(const Rectangle &r) const
{
	return (r.br >= tl && r.tl <= br);
}
#endif

#ifdef RECTM
Rectangle& Rectangle::operator&=(const Rectangle &r)
{
	ASSERT(this->intersects(r));
	if(r.encloses(*this))
		;
	else if(this->encloses(r))
		*this = r;
	else
		build(max(this->tl.x,r.tl.x),
				max(this->tl.y,r.tl.y),
				min(this->br.x,r.br.x),
				min(this->br.y,r.br.y));
	return *this;
}
#endif

#ifdef RECTN
Rectangle& Rectangle::operator|=(const Rectangle &r)
{
	if(r.encloses(*this))
		*this = r;
	else if(this->encloses(r))
		;
	else
		build(min(this->tl.x,r.tl.x),
				min(this->tl.y,r.tl.y),
				max(this->br.x,r.br.x),
				max(this->br.y,r.br.y));
	return *this;
}
#endif
