// Basic Screen class.
// All drawing on the screen is through a Screen.
// The standard implementation uses the whole screen as the only Screen.
#ifndef SCREEN_HPP
#define SCREEN_HPP

#ifndef RECT_HPP
#include "rect.hpp"
#endif

/*
This class is used as the interface to the screen hardware. Most of
its methods are private (Window is a friend, so can call them). You
would only override this class if you wanted to implement the
toolkit on different hardware, or improve the interface. A screen
pointer variable is provided to access the hardware screen (e.g. to
obtain its size). 
*/
class Screen : private Rectangle {
public:
	coord width() const { return Rectangle::width(); }
	coord height() const { return Rectangle::height(); }
	const Rectangle& rectangle() const { return (Rectangle &)*this; }
	Rectangle::inspectinfo;
	void emergencyMessage(const char *s);
	// This method is provided to allow an emergence close-down message to
	// be written to the screen without any memory allocation or DOS
	// functions being called. It is provided so that a new_handler can
	// exit cleanly if the system runs out of memory. 
private:
friend class Window;
friend class Application;
	Point cursorpt;
	int cursorHidden;

	Screen();
	~Screen();

// scroll a rectangular area
// set nlines = 0 to clear
// set nlines < 0 to scroll down instead of up
	void scroll(const Rectangle &a,const color c,int nlines = 1);

// print a string

	void printstring(Point at,const color c,const char *s,int length);

// print a string, space filling on right to length average characters

	void fprintstring(Point at,const color c,const char *s,const int length);

// fill a rectangle with a character

	void fillrect(const Rectangle& r,const color c,const char x);

// Move text cursor to p

	void cursor(const Point& p);

// show and hide text cursor

	void showcursor(int block = 0);
	void hidecursor();

// update physical screen to reflect latest state of Screen
	void update();

	int enable();
	void disable();
	int iscolor() const { return colorflag; }  // 1 if a color screen
private:
	int colorflag;
	};

// The actual physical Screen
extern Screen *screen;

#endif

