#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <dos.h>
#ifdef disable      /* TURBO C #defines enable & disable in dos.h ! */
#undef disable
#endif
#ifdef enable
#undef enable
#endif
#include "applic.hpp"
#include "editclus.hpp"
#include "fileedit.hpp"
#include "timeedit.hpp"
#include "menu.hpp"
#include "panel.hpp"
#include "text.hpp"
#include "alert.hpp"
#include "dialog.hpp"
#include "radio.hpp"
#include "check.hpp"
#include "event.hpp"
#include "colset.hpp"

EventQueue eventQueue;
Application app;

FileList *flist;

class SortButton : public ActionButton {
public:
    SortButton(int (*compare)(const FoundFile *,const FoundFile *),
        const char *name,Keys hotkey = Knone);
    virtual int doit();
private:
    int (*order)(const FoundFile *,const FoundFile *);
    };

SortButton::SortButton(int (*compare)(const FoundFile *,const FoundFile *),
    const char *name,Keys hotkey)
: ActionButton(0,name,hotkey), order(compare)
{
}

int SortButton::doit()
{
    flist->sort(order);
    return 1;
}

static char dirname[OSPathLen+1];

char *expanddir(const char *s)
{
    static char newname[OSPathLen+1];

    strcpy(newname,s);
    fixdirname(newname);
    return newname;
}

UpperEdit *directory;

int sfiles()
{
    static char newname[OSPathLen+1];

    directory->setString(dirname);
    strcpy(newname,fullfilename(dirname,"*.*"));
    flist->deleteall();
    flist->addfiles(newname);
    return flist->count();
}

int selectfile(const char *)
{
    static char oldname[OSPathLen+1];
    const FoundFile *f = flist->selected();

    strcpy(oldname,dirname);
    if(!f || !(f->attribute(FAdirectory)) || strcmp(directory->string(),dirname))
        strcpy(dirname,expanddir(directory->string()));
    else
        strcpy(dirname,expanddir(fullfilename(dirname,f->fname())));
    if(!sfiles())
        {
        alert(form("No files in directory\n%s",dirname));
        strcpy(dirname,oldname);
        sfiles();
        }
    return 1;
}

Text *opttext;

int selectOpt(const char *o)
{
    opttext->setString(o);
    return 1;
}

int doDialog(const char *)
{
    static int i = 0, r = 0;
    static unsigned u = 0, chk = 0;
    static char fname[35] = "*.*";
    static Time time;

    EditVCluster *e = new EditVCluster(GCollection(EditControl)(3)
    	+ new EditCluster(GCollection(EditControl)(8)
	        + new TextEdit("Test 1",10,"TextEdit")
	        + new UpperEdit("UPPER",10,"UpperEdit")
	        + new LowerEdit("lower",10,"LowerEdit")
	        + new UnsignedEdit(u,"UnsignedEdit")
	        + new IntEdit(i,"IntEdit")
	        + new DateEdit(time,"Date")
	        + new TimeEdit(time,"Time")
	        + new FileEdit(fname,"FileEdit"),
	        "Test TextEdit",AlignFlag(AFHleft|AFVtop),
	        	BorderFlag(BFsingle|BFtitle)
		    )
    	+ new RadioHCluster(r,GCollection(constchar)(3)
	        + "Radio~1"
	        + "Radio~2"
	        + "Radio~3",
	        "Radio Buttons",BorderFlag(BFtitle|BFsingle|BFtabstop|BFleft)
		    )
	    + new EditHCluster(GCollection(EditControl)(3)
	    	+ new CheckButton("Check ~1",chk,BIT(0))
	        + new CheckButton("Check ~2",chk,BIT(1))
	        + new CheckButton("Check ~3",chk,BIT(2)),
            "Check buttons",AlignFlag(AFHexpand|AFtabstop), 
            BorderFlag(BFsingle|BFtitle|BFright)
	        )
	    );

    Dialog *d = new Dialog(
        new VCluster(ObjectCollection(2) 
            + e
            + new HCluster(ObjectCollection(4)
                + new CloseEditButton(e)
                + new ApplyEditButton(e)
                + new ResetEditButton(e)
                + new CancelButton,
                AlignFlag(AFHexpand|AFtabstop))
            ),
        "Test Dialog Box");

    d->enter();
    return 1;
}

Panel *mainPanel;

extern ColorSet alertColors, helpColors;

int setDefaultColor(const char *)
{
	if(!setColorDialog(defaultColors))
		mainPanel->setcolors(defaultColors);
	return 1;
}

int setAlertColor(const char *)
{
	setColorDialog(alertColors);
	return 1;
}

int setHelpColor(const char *)
{
	setColorDialog(helpColors);
	return 1;
}

int crashnow(const char *)
{
	crash("Test crash", __FILE__, __LINE__);
	return 1;
}

main(int argc,char **argv)
{
#ifndef NDEBUG
    FUNCTION("main");

    if(argc > 1)
        {
		if(argv[1][0] == '-')
			VObject::debug = stderr;
		else
			{
		    VObject::debug = fopen(argv[1],"w");
		    static char buffer[512];
        	setvbuf(VObject::debug,buffer,_IOLBF,512);
			}
        }
#endif

	int toggle = 0;
    flist = new FileList("*.*");
    strcpy(dirname,expanddir("."));
    directory = new UpperEdit(dirname,OSPathLen,"Directory");
    opttext = new Text("         ");
    mainPanel = new Panel(
        new MenuBar(KF10,OCollection(MenuButton)(4)
            + new MenuButton("Menu ~A",Kauto,LCollection(Option)(5)
                + new Option("Test ~Dialog",doDialog)
                + new TickOption("~Toggle",toggle)
                + new Option("Option ~1",selectOpt,KAltF1)
                + new Option("Option ~2",selectOpt,KAltF2)
                + new Option("Option ~3",selectOpt,KAltF3)
                )
            + new MenuButton("Menu ~B",Kauto,LCollection(Option)(5)
                + new Option("Test ~Dialog",doDialog)
                + new TickOption("~Toggle",toggle)
                + new Option("Option ~4",selectOpt,KAltF4)
                + new Option("Option ~5",selectOpt,KAltF5)
                + new Option("Option ~6",selectOpt,KAltF6)
                )
            + new MenuButton("Menu ~C",Kauto,LCollection(Option)(5)
                + new Option("Test ~Dialog",doDialog)
                + new TickOption("~Toggle",toggle)
                + new Option("Option ~7",selectOpt,KAltF7)
                + new Option("Option ~8",selectOpt,KAltF8)
                + new Option("Option ~9",selectOpt,KAltF9)
                )
            + new MenuButton("~Set Colors",Kauto,LCollection(Option)(4)
                + new Option("~Default Colours",setDefaultColor)
                + new Option("~Alert Colors",setAlertColor)
                + new Option("~Help Colors",setHelpColor)
                + new Option("Crash", crashnow)
                )
            + DefaultHelpMenu
            ),
        new HCluster(ObjectCollection(3)
            + new VCluster(ObjectCollection(4)
                + directory
                + new HCluster(ObjectCollection(1)
                    + new ActiveButton(new FunctionCommand(selectfile),
						"New Directory",KAltD,Default),
                    AlignFlag(AFexpand|AFtabstop))
                + flist
                + new HCluster(ObjectCollection(4)
                    + new SortButton(sortname,"Name",KAltN)
                    + new SortButton(sortext,"Ext",KAltE)
                    + new SortButton(sorttime,"Time",KAltT)
                    + new SortButton(sortsize,"Size",KAltS),
                    "Sequence",AFexpand,BorderFlag(BFsingle|BFtitle|BFtabstop))
                )
            + new VCluster(ObjectCollection(4)
                + new Text("Top line")
                + opttext
                + new MultiLineText("Multiple lines\nof text (line 2)\nline 3")
                + new Text("Bottom line")
                )
            + new VCluster(ObjectCollection(3)
                + new ActiveButton(0,"Active Button")
                + new CloseButton(0,notDefault)
                + new CancelButton,
                AlignFlag(AFexpand|BFtabstop))
            )
        );
    mainPanel->show();
    Application::openHelpFile("T");
    CriticalErrorHandler catchErrors;
    app.enter();
    return 0;
}
