// Text items
#ifndef TEXT_HPP
#define TEXT_HPP
#ifndef VOBJECT_HPP
#include "vobject.hpp"
#endif

/*
A simple displayed string. The text is stored in an internal buffer 
allocated for the purpose.
*/
class Text : public VObject {
public:
	Text(const char *string = "",int length = -1);
	// The default length is the length of the string.
	~Text();

	void setString(const char *string);
	// Change the string. The original length will be used (the
	// string will be truncated to fit if necessary).
	const char *string() const { return text; }
	// The current string.
	int maxlen() const { return textlen; }
	// The current maximum string length.

	virtual coord minwidth() const;
	virtual void paint(Rectangle r);
	virtual char *inspectinfo(char *buffer = 0,const char *s = 0) const;
	virtual char *debuginfo(char *buffer = 0) const;

protected:
//~ Constructors for subclasses where size not dependent on text length
	Text(const Point &size,const char *string = "",int length = -1);
	Text(coord width,coord height,const char *string = "",int length = -1);

private:
	void build(const char *string,int length);
	char *text;
	int textlen;
	};

// A text string that centers itself.
class CenterText : public Text {
public:
	CenterText(const char *string,int length = -1);
	// If length is not specified, the length of the string is used as the
	// initial size of the CenterText. This is pretty pointless unless the
	// object is later enlarged (perhaps by including in a Cluster).
	virtual void paint(Rectangle r);
	};

// The text is right justified in the rectangle.
class RightText : public Text {
public:
	RightText(const char *string,int length = -1);
	virtual void paint(Rectangle r);
	};

// Text containing embedded newlines. Tab characters are not 
// expanded.
class MultiLineText : public Text {
public:
	MultiLineText(const char *string = "",int length = -1);
	// The string is printed in a rectangle large enough to hold it. The
	// default length is the length of the string - shorter lengths will
	// only print part of the string. Longer lengths might be necessary if
	// you intend to change the string later. 
	MultiLineText(coord width,coord height,const char *string = "",
		int length = -1);
	// The rectangle is specified explicitly.

	virtual coord minwidth() const;
	virtual coord minheight() const;

// draw the Object
	virtual void paint(Rectangle r);
	};

/*
This is the "cheapest" way of getting fixed, static text displayed
on the screen. No memory is allocated to hold a copy of the text
string supplied, so it had better be a constant string. 
*/
class FixedText : public VObject {
public:
	FixedText(const char *string);

	virtual coord minwidth() const;
	virtual void paint(Rectangle r);
	};

#endif
