#ifdef ALL
#define TEXTEDA
#define TEXTEDB
#define TEXTEDC
#define TEXTEDD
#define TEXTEDE
#define TEXTEDF
#endif
#ifndef TEXTEDA
#define LIBRARY_MODULE
#endif

#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <ctype.h>
#include "textedit.hpp"
#include "popwin.hpp"
#include "event.hpp"
#include "screen.hpp"

#ifdef TEXTEDA
implement(OCollection,TextEdit)

TextEdit::TextEdit(char *string,int length,const char *prompt,
	const char *validchars)
: Control(length,1,prompt), parameter(string), browseWindow(0), 
    pos(0), textlen(length), eflags(EFchanged), validation(validchars),
    offset(0), selected(0)
{
    setFlag(StateFlag(SFshowfocus|SFtabstop));
    text = new char[maxlen()+1];
    edittext = new char[maxlen()+1];
    ASSERT(text && edittext);
    setString(string);
	int lcaseok = 0, ucaseok = 0;
	while(*validchars)
		{
		if(validchars[1] == '-')
			{
			if(*validchars <= 'Z' && validchars[2] >= 'A')
				ucaseok = 1;
			if(*validchars <= 'z' && validchars[2] >= 'a')
				lcaseok = 1;
			validchars += 3;
			}
		else
			{
			if(*validchars <= 'Z' && *validchars >= 'A')
				ucaseok = 1;
			if(*validchars <= 'z' && *validchars >= 'a')
				lcaseok = 1;
			validchars++;
			}
		}
	if(lcaseok && !ucaseok)
		setEflag(EFlcase);
	else if(ucaseok && !lcaseok)
		setEflag(EFucase);
}

TextEdit::~TextEdit()
{
    delete text;
    delete edittext;
}

void TextEdit::setString(const char *string)
{
    if(maxlen() > 0)
        strncpy(text,string,maxlen());
    text[maxlen()] = 0;
    setChanged();
    dirty();
}

coord TextEdit::maxwidth() const
{
    return canBrowse() ? textlen : textlen + 1;
}

coord TextEdit::minwidth() const
{
    return 5;
}

void TextEdit::paint(Rectangle r)
{
    int color = getcolor();

	if(selected && pos == 0)
		color |= SFhighlight;
    Control::paint(r);
    window->printstring(topleft(),r,color,string()+offset);
    if(canBrowse())
        {
        Point p = { right(), top() };
        window->printstring(p,r,getcolor()^SFhighlight,DownArrow);
        }
}

int TextEdit::leftclick(MouseEvent& e)
{
    setfocus();
    selected = 0;
    if(e.at().x == right() && e.at().y == top() && canBrowse())
        {
        VObject *b = browse();
        if(b)
            startBrowse(b);
        }
    else
        {
        pos = e.at().x - left();
        if(pos > strlen(string()))
            pos = strlen(string());
        updateCursor();
        }
    return 1;
}

int TextEdit::kbdevent(KeyboardEvent& e)
{
    int check = 0, oldpos = pos;
    
    switch (e.value())
        {
    case Kleft:
        if(pos)
            pos--;
        break;
    case Kright:
        if(string()[pos])
            pos++;
        break;
    case Khome:
        pos = 0;
        break;
    case Kend:
        pos = strlen(string());
        break;
    case Kinsert:
        toggleInsert();
        window->showcursor(insmode());
        break;
    case KBksp:
        if(!pos)
            break;
        pos--;      // fall through: backspace is equivalent to left+del.
        if(!insmode() && string()[pos + 1])
            break;  // overtype mode backspace non-destructive except at end
    case Kdel:
        strcpy(edittext,string());
        delchars(pos);  // delete the character at pos
        check = 1;
        break;
    case Kctrlright:
        while(string()[pos])
            if(isspace(string()[pos++]) && !isspace(string()[pos]))
                break;
        break;
    case Kctrlleft:
        while(pos && --pos)
            if(isspace(string()[pos]) && !isspace(string()[pos-1]))
                break;
        break;
    case Kctrlhome:
        if(pos)
            {
            strcpy(edittext,string());
            delchars(0,pos);
            pos = 0;
            check = 1;
            }
        break;
    case Kctrlend:
        if(string()[pos])
            {
            strcpy(edittext,string());
            delchars(pos,strlen(string()+pos));
            check = 1;
            }
        break;
	case ctrl('R'):				// Restore orig value
		read();
		pos = 0;
		updateCursor();
		break;
    case KF4:                   // browse
    case Kdown:
        if(canBrowse())
            {
            VObject *b = browse();
            if(b)
                startBrowse(b);
            return 1;
            }
        return 0;
    default:
        if(e.value() >= 256)
            return Control::kbdevent(e);
        strcpy(edittext,string());

        char c = filter(e.value());
        if(!c)
            return 0;
        if(pos >= maxlen())
            break;

		if(selected && !pos)
			edittext[0] = 0;
        if(insmode())
            if(!inschars(pos))
                break;
        if(!edittext[pos])
            edittext[pos + 1] = 0;          // ensure correctly terminated
        edittext[pos++] = c;
        check = 1;
        }
    if(check)
        {
        if(validate())
            updateString();
        else
            pos = oldpos;
        }
    if(pos != oldpos)
        updateCursor();
	if(selected)
		{
		selected = 0;
		dirty();
		}
    return 1;
}

int TextEdit::setfocus(int tabstop)
{
    int l = strlen(string());

    if(pos > l)
        pos = l;
    if(testcFlag(CFfocus))
        {
        updateCursor();
        return 1;
        }
    if(!Control::setfocus(tabstop))         // bypass Text setfocus
        return 0;
	offset = 0;
	selected = 1;
    updateCursor();
    if(browseWindow)
        browseWindow = 0;
    else
        {
        if(defaultMode)
            insertOn();
        else
            insertOff();
        }
    window->showcursor(insmode());
    return 1;
}

void TextEdit::resetfocus()
{
    if(!testcFlag(CFfocus))
        return;
    if(hasChanged() && !browseWindow)
        valid();
    offset = pos = 0;
    selected = 0;
    window->hidecursor();
    Control::resetfocus();
}

void TextEdit::move(coord x,coord y)
{
    Control::move(x,y);
    if(testcFlag(CFfocus))
        updateCursor();
}

char TextEdit::filter(char inchar)
{
	if(!*validation)
	    return iscntrl(inchar) ? 0 : inchar;    // filter control chars
	if(testEflag(EFucase))
		inchar = toupper(inchar);
	else if(testEflag(EFlcase))
		inchar = tolower(inchar);

	const char *p = validation; 

	while(*p)
		{
		if(p[1] == '-')
			{
			if(*p <= inchar && p[2] >= inchar)
				return inchar;
			p += 3;
			}
		else if(inchar == *p++)
			return inchar;
		}
	return 0;
}

int TextEdit::validate()
{
    return 1;                   // default is accept anything
}

#ifdef __TURBOC__
TextEdit::
#endif
Complete TextEdit::complete()
{
    return COK;                 // default is accept anything
}

int TextEdit::valid()
{
    if(!hasChanged())
        return (isValid() != 0);
    strcpy(edittext,string());
    if(!validate())     // double check we have a valid string
        {
        setInvalid();
        return 0;
        }
    switch(complete())
        {
    case COK:
    case Cchanged:
        break;
    case Cinvalid:
        setInvalid();
        return 0;
        }
    updateString();
    setValid();
    setUnchanged();
    return 1;
}

void TextEdit::read()
{
    setString(parameter);
}

int TextEdit::writeback()
{
    strcpy(parameter,string());
    return 1;
}

int TextEdit::defaultMode = 0;  // default mode is overtype

int TextEdit::inschars(int posn,int nchars)
{
    if(!nchars)
        return 1;
    int l = strlen(edittext);
    if(nchars < 0 || posn < 0 || posn > l || l + nchars > maxlen())
        return 0;
    memmove(edittext+posn+nchars,edittext+posn,maxlen()-posn-nchars+1);
    edittext[posn] = ' ';
    return 1;
}

int TextEdit::delchars(int posn,int nchars)
{
    if(!nchars)
        return 1;
    int l = strlen(edittext);
    if(nchars < 0 || posn < 0 || posn > l - nchars)
        return 0;
    memmove(edittext+posn,edittext+posn+nchars,maxlen()-posn-nchars+1);
    return 1;
}

VObject *TextEdit::browse()
{
    return 0;
}

void TextEdit::updateCursor()
{
	if(pos < offset)
		{
		offset = pos;
		dirty();
		}
	else if(pos > offset + width() - 1)
		{
		offset = pos - width() + 1;
		dirty();
		}
    window->cursor(left()+pos-offset,top());
}

void TextEdit::startBrowse(VObject *v)
{
    browseWindow = new PopWindow(v);
    browseWindow->popin(this);
}

VObject *TextEdit::inheritingVObject()
{
	return this;
}
#endif

#ifdef TEXTEDB
void TextEdit::endBrowse(const char *chosen)
{
    if(chosen)
        {
	    if(maxlen() > 0)
	        strncpy(edittext,chosen,maxlen());
	    edittext[maxlen()] = 0;
        if(validate())
            updateString();
        }
    else
        read();
    if(hasChanged() && !testcFlag(CFfocus))
        valid();
    if(browseWindow)
		{
		VALIDATE(browseWindow);
        delete browseWindow;
		}
}
#endif

#ifdef TEXTEDC
UpperEdit::UpperEdit(char *string,int length,const char *prompt,
	const char *validchars)
: TextEdit(string,length,prompt,validchars)
{
	setCase(EFucase);
}
#endif

#ifdef TEXTEDD
LowerEdit::LowerEdit(char *string,int length,const char *prompt,
	const char *validchars)
: TextEdit(string,length,prompt,validchars)
{
	setCase(EFlcase);
}
#endif

#ifdef TEXTEDE
UnsignedEdit::UnsignedEdit(unsigned &i,const char *prompt,
    unsigned low,unsigned high)
: TextEdit(form("%u",i),5,prompt,"0-9"), uparameter(i), lowval(low), 
	highval(high)
{
    unsigned long powerOfTen = 1;

    for(int j = 1; powerOfTen < UINT_MAX; j++, powerOfTen *= 10)
        if(high < powerOfTen)
            break;
    setwidth(j);
}

#ifdef __TURBOC__
TextEdit::
#endif
Complete UnsignedEdit::complete()
{
    unsigned long l = atol(edittext);

    return (l <= highval && l >= lowval) ? COK : Cinvalid;
}

void UnsignedEdit::read()
{
    setString(form("%d",uparameter));
}

int UnsignedEdit::writeback()
{
    uparameter = atol(string());
    return 1;
}
#endif

#ifdef TEXTEDF
IntEdit::IntEdit(int &i,const char *prompt,int low,int high)
: TextEdit(form("%d",i),6,prompt,"-0-9"), iparameter(i), lowval(low), 
	highval(high)
{
    long powerOfTen = 1;

    for(int j = 1; powerOfTen < INT_MAX; j++, powerOfTen *= 10)
        if(low > -powerOfTen && high < powerOfTen)
            break;
    setwidth(j + (low < 0));
}

char IntEdit::filter(char inchar)
{
    return (isdigit(inchar) || (!pos && inchar == '-')) ? inchar : 0;
}

int IntEdit::validate()
{
    char *s = strchr(edittext,'-');

    if(s && s != edittext)
        return 0;
    return 1;
}

#ifdef __TURBOC__
TextEdit::
#endif
Complete IntEdit::complete()
{
    long l = atol(edittext);

    return (l <= highval && l >= lowval) ? COK : Cinvalid;
}

void IntEdit::read()
{
    setString(form("%d",iparameter));
}

int IntEdit::writeback()
{
    iparameter = atol(string());
    return 1;
}
#endif
