#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <ctype.h>
#include "time.hpp"

#ifdef ALL
#define TIMEA
#define TIMEB
#define TIMEC
#define TIMED
#define TIMEE
#define TIMEF
#define TIMEG
#define TIMEH
#define TIMEI
#define TIMEJ
#define TIMEK
#define TIMEL
#define TIMEM
#define TIMEN
#define TIMEO
#define TIMEP
#define TIMEQ
#define TIMER
#define TIMES
#define TIMET
#define TIMEU
#define TIMEV
#define TIMEW
#define TIMEX
#define TIMEY
#define TIMEZ
#endif

#define TIME_FUDGE (7 * 60 * 60)

extern int get3values(const char *string,unsigned *values);

#ifdef TIMEA
Time::Time(long daysSinceJan1_1970)
: tim(daysSinceJan1_1970 + TIME_FUDGE)
{
}
#endif

#ifdef TIMEB
Time::Time()
: tim(::time(&tim))
{
}
#endif

#ifdef TIMEC
Time::Time(const Time &t)
: tim(t.tim)
{
}
#endif

#ifdef TIMED
Time::Time(unsigned dosdate, unsigned dostime)
: tim(0)
{
	setdostime(dosdate,dostime);
}
#endif

#ifdef TIMEE
Time::Time(const char *filename)
:tim(0)
{
	filetime(filename);
}
#endif

#ifdef TIMEF
int Time::setdate(unsigned int d, unsigned int m, unsigned int y)
{
	struct tm temptime;
	
	temptime = *localtime(&tim);
	temptime.tm_mday = d;
	temptime.tm_mon = m - 1;
	temptime.tm_year = y - 1900;

	long newtime = mktime(&temptime);
	if(newtime >= 0
	&& temptime.tm_mday == d
	&& temptime.tm_mon == m - 1
	&& temptime.tm_year == y - 1900)
		{
		tim = newtime;
		return 1;
		}
	return 0;
}
#endif

#ifdef TIMEG
static Time zTime(0L);
Time &Time::zero = zTime;

int get3values(const char *string,unsigned *values)
{
	int p = 0, found = 0;

	values[0] = values[1] = values[2] = 0;
	for(int n = 0; n < 3 && *string; string++)
		{
		if(isdigit(*string))
			{
			values[n] = 10 * values[n] + *string - '0';
			found = n + 1;
			if(!p++ || (n == 2 && p < 4) || ispunct(string[1]))
				continue;
			}
		else if(!ispunct(*string))
			return 0;
		n++;
		p = 0;
		}
	if(*string)
		found = 0;
	return found;
}
#endif

#ifdef TIMEH
int Time::setdate(const char *string)
{
	unsigned int values[3];
	int found = get3values(string,values);

	if(!found)
		return 0;
	if(found < 3)
		{
		values[2] = year();
		if(found < 2)
			values[1] = month();
		}
	else if(values[2] < 50)
		values[2] += 2000;
	else if(values[2] < 100)
		values[2] += 1900;
	return setdate(values[0],values[1],values[2]);
}
#endif

#ifdef TIMEI
void Time::setdate(unsigned daysSinceJan1_1970)
{
	tim = daysSinceJan1_1970 * 24L * 60 * 60 + TIME_FUDGE + secsSinceMidnight();
}
#endif

#ifdef TIMEJ
int Time::settime(unsigned int h, unsigned int m, unsigned int s)
{
	struct tm temptime;
	
	if(h >= 24 || m >= 60 || s >= 60)
		return 0;
	temptime = *localtime(&tim);
	temptime.tm_hour = h;
	temptime.tm_min = m;
	temptime.tm_sec = s;

	long newtime = mktime(&temptime);
	if(newtime >= 0)
		{
		tim = newtime;
		return 1;
		}
	return 0;
}
#endif

#ifdef TIMEK
int Time::settime(long secs)
{
	if(secs >= 24L * 60 * 60)
		return 0;

	tim -= secsSinceMidnight();
	tim += secs;
	return 1;
}
#endif

#ifdef TIMEL
int Time::settime(const char *string)
{
	unsigned int values[3];
	int found = get3values(string,values);

	if(!found)
		return 0;
	if(found < 3)
		{
		values[2] = second();
		if(found < 2)
			values[1] = minute();
		}
	return settime(values[0],values[1],values[2]);
}
#endif

#ifdef TIMEM
void Time::systemtime()
{
	::time(&tim);
}
#endif

#ifdef TIMEN
int Time::setdostime(unsigned date,unsigned time)
{
	struct tm temptime;
	
	temptime.tm_hour = (time >> 11) & 0x1f;
	temptime.tm_min = (time >> 5) & 0x3f;
	temptime.tm_sec = (time << 1) & 0x3f;
	temptime.tm_year = ((date >> 9) & 0x7f) + 80;
	temptime.tm_mon = ((date >> 5) & 0x0f) - 1;
	temptime.tm_mday = date & 0x1f;

	long newtime = mktime(&temptime);

	if(newtime >= 0)
		{
		tim = newtime;
		return 1;
		}
	return 0;
}
#endif

#ifdef TIMEO
int Time::filetime(const char *filename)
{
	struct stat st;

	if(stat((char *)filename,&st))
		return 0;
	tim = st.st_mtime;
	return 1;
}
#endif

#ifdef TIMEP
unsigned int Time::dayno() const
{
	return (tim - TIME_FUDGE) / (24L * 60 * 60);
}
#endif

#ifdef TIMEQ
long Time::secsSinceMidnight() const
{
	return (tim - TIME_FUDGE) % (24L * 60 * 60);
}
#endif

#ifdef TIMER
unsigned Time::dostime() const
{
	struct tm *tt = localtime(&tim);

	unsigned t = tt->tm_hour;
	t = (t << 6) | tt->tm_min;
	t = (t << 5) | (tt->tm_sec >> 1);
	return t;
}
#endif

#ifdef TIMES
unsigned Time::dosdate() const
{
	struct tm *tt = localtime(&tim);

	unsigned d = tt->tm_year - 80;
	d = (d << 4) | (tt->tm_mon + 1);
	d = (d << 5) | (tt->tm_mday);
	return d;
}
#endif

#ifdef TIMET
unsigned Time::hour() const
{
	return localtime(&tim)->tm_hour;
}
#endif

#ifdef TIMEU
unsigned Time::minute() const
{
	return localtime(&tim)->tm_min;
}
#endif

#ifdef TIMEV
unsigned Time::second() const
{
	return localtime(&tim)->tm_sec;
}
#endif

#ifdef TIMEW
unsigned Time::year() const
{
	return localtime(&tim)->tm_year + 1900;
}
#endif

#ifdef TIMEX
unsigned Time::month() const
{
	return localtime(&tim)->tm_mon + 1;
}
#endif

#ifdef TIMEY
unsigned Time::day() const
{
	return localtime(&tim)->tm_mday;
}
#endif

#ifdef TIMEZ
extern "C" char *form(const char *format,...);
const char *Time::formattime(const char *format)
{
	static char timebuf[81];

	if(strftime(timebuf,80,format,localtime(&tim)))
		return form("%s",timebuf);
	return "";
}
#endif
