#ifdef ALL
#define WINDOWA
#define WINDOWB
#define WINDOWC
#define WINDOWD
#define WINDOWE
#define WINDOWF
#define WINDOWG
#define WINDOWH
#define WINDOWI
#define WINDOWJ
#define WINDOWK
#define WINDOWL
#define WINDOWM
#define WINDOWN
#endif
#ifndef WINDOWA
#define LIBRARY_MODULE
#endif

#include <string.h>
#include "applic.hpp"               // includes window.hpp
#include "screen.hpp"
#include "command.hpp"
#include "event.hpp"

// This class maintains a list of clipping rectangles

class ClipRectangle : public Rectangle {
public:
    ClipRectangle(class GStack(ClipRectangle)& list,const Rectangle& r);
// WARNING - unlike Rectangle, bx and by are coordinates of bottom left
// corner, not width & height. This is for convenience when calling.
    ClipRectangle(class GStack(ClipRectangle)& list,const Point& p,const coord bx,const coord by);
    };

declare(GStack,ClipRectangle);

#ifdef WINDOWA
ClipRectangle::ClipRectangle(GStack(ClipRectangle)& list,const Rectangle& r)
: Rectangle(r)
{
    list.push(this);
}

ClipRectangle::ClipRectangle(GStack(ClipRectangle)& list,const Point& p,
    const coord bx,const coord by)
: Rectangle(p,bx-p.x+1,by-p.y+1)
{
    list.push(this);
}

static GStack(ClipRectangle) cliplist;  // Collection of clipping Rectangles

inline ClipRectangle *Window::cliprect() const
{
    return cliplist.pop();
}

void Window::startclip(Rectangle &clip) const
{
    if(!this->intersects(clip))         //result - outside window
        return;

    clip &= *this;

    new ClipRectangle(cliplist,clip);   //add first rectangle to cliplist

    ObjectIterator next(*application);
    for(Window *w = (Window *)next(); w != this; w = (Window *)next())
        {
        ASSERT(w /* Window not in application */);

        GStack(ClipRectangle) list;

        for(ClipRectangle *r = cliplist.pop(); r; 
          r = cliplist.pop())
            {
            w->split(list,*r);
            delete(r);
            }
        cliplist = list;
        }
}

// print a string (can be on any Window)
void Window::printstring(Point at,const Rectangle &clip,int color,
    const char *s,int length)
{
    setCurrentColorFromState(color);
    if(length < 0)
        length = strlen(s);
    if(!length)
        return;
    Rectangle r(at,length,1);
    if(!r.intersects(clip))
        return;
    r &= clip;
    screenCoords(at);
    screenCoords(r);

    ClipRectangle *cl;

    startclip(r);
    while((cl = cliprect()) != 0)
        {
        screen->printstring(cl->topleft(),currentColor,
            s+cl->left()-at.x,cl->width());
        delete cl;
        }
}

void Window::printstring(const Point &at,int color,const char *s,int length)
{
    printstring(at,*screen,color,s,length);
}

void Window::fillrect(Rectangle r,int color,const char x)
{
    setCurrentColorFromState(color);
    screenCoords(r);
    startclip(r);

    ClipRectangle *clip;

    while((clip = cliprect()) != 0)
        {
        screen->fillrect(*clip,currentColor,x);
        delete clip;
        }
}

void Window::clear(int color,const char x)
{
    fillrect(Rectangle(origin,width(),height()),color,x);
}

void Window::split(GStack(ClipRectangle)& list,ClipRectangle& clip) const
{
    if(this->encloses(clip))
        return;

    if(!this->intersects(clip))
        {
        new ClipRectangle(list,clip);
        return;
        }

    Point p = clip.topleft(), q = clip.botright();

    if(this->top() > clip.top())
        {
        // process first third y
        new ClipRectangle(list,p,q.x,this->top() - 1);
        p.y = this->top();
        }
        // process middle third
    if(clip.bottom() > this->bottom())
        q.y = this->bottom();
    if(this->left() > clip.left())
        // process left section
        new ClipRectangle(list,p,this->left() - 1,q.y);
    if(this->right() < clip.right())
        {
        // process right section
        p.x = this->right() + 1;
        new ClipRectangle(list,p,q.x,q.y);
        p.x = clip.left();
        }
    p.y = this->bottom() + 1;
    if(this->bottom() < clip.bottom())
        // process last third y
        new ClipRectangle(list,p,clip.right(),clip.bottom());
}
#endif

#ifdef WINDOWB
Window::Window(const char *n,const ColorSet &c)
: Group(n), currentColor(c.colors[0]), colors(c), actfocus(0)
{
    cursorpt.x = cursorpt.y = 0;
}
#endif

#ifdef WINDOWC
Window::Window(VObject *v,AlignFlag a,const ColorSet &c,const char *n)
: Group(n), currentColor(c.colors[0]), colors(c), actfocus(0)
{
    cursorpt.x = cursorpt.y = 0;
    if(v)
        setsize(v->width(),v->height());
    align(a,*screen);
    if(v)
        {
        v->setsize(width(),height());
        add(v);
        }
}
#endif

#ifdef WINDOWD
Window::Window(const coord l,const coord t,const coord r,const coord b,
    VObject *v,const ColorSet &c,const char *n)
: Group(l,t,r,b,n), currentColor(c.colors[0]), colors(c), actfocus(0)
{
    cursorpt.x = l;
    cursorpt.y = t;
    if(v)
        {
        v->setsize(width(),height());
        add(v);
        }
}
#endif

#ifdef WINDOWE
Window::~Window()
{
    if(application)
        application->removeWindow(this);
}

void Window::contentsDeleted(VObject* v)
{
    Rectangle r(*v);
    if(objects.remove(v))
        {
        screenCoords(r);
        dirty(r);
        }
}

int Window::closeok()
{
    new CommandEvent(new CloseCommand(this));
    return 1;       // close is OK by default
}

int Window::cancelok()
{
    new CommandEvent(new CloseCommand(this));
    return 1;       // cancel is OK by default
}

void Window::move(coord x,coord y)
{
    VObject::move(x,y);     // don't move the contents as their
                            // positions are relative to this Window
}

void Window::paint(Rectangle r)
{
    r &= *this;
#ifndef NDEBUG
    if(debug)
        fprintf(debug,"%s->paint%s\n",debuginfo(),r.inspectinfo());
#endif
    windowCoords(r);
    fillrect(r,getcolor(),' ');
    paintcontents(r);
}

int Window::kbdevent(KeyboardEvent& e)
{
    int r;

    if(actfocus)
        {
        r = actfocus->kbdevent(e);
        if(r)
            return r;
        }
    if(e.value() == KTab || e.value() == KBackTab)
        {
        if(nextTabstop(e.value() == KBackTab,1))
            return 1;
        }
    return Group::kbdevent(e);
}

#ifdef NINSPECT
char *Window::inspectinfo(char *buf,const char *s) const
{
    return "";
}
#else
char *Window::inspectinfo(char *buf,const char *s) const
{
    if(!s)
        s = "Window ";
    return Group::inspectinfo(buf,s);
}
#endif
#endif

#ifdef WINDOWF
// Manipulate Windows on screen
void Window::enter()
{
    application->removeWindow(this);
    application->pop(this);
}
#endif

#ifdef WINDOWG
void Window::show()
{
    if(this == application->first())
        {
        dirty();
        return;
        }
    application->removeWindow(this);
    application->addWindow(this);
}
#endif

#ifdef WINDOWH
void Window::removeWindow()
{
    application->removeWindow(this);
}
#endif

#ifdef WINDOWI
void Window::setactive(VObject *v)
{
    if(actfocus && actfocus != v)
        {
        actfocus->clearFlag(SFactfocus);
        actfocus->dirty();
        }
    if(actfocus != v)
        {
        actfocus = v;
        actfocus->setFlag(SFactfocus);
        actfocus->dirty();
        }
}
#endif

#ifdef WINDOWJ
void Window::clearactive(VObject *v)
{
    if(v && actfocus == v)
        {
        actfocus->clearFlag(SFactfocus);
        actfocus->dirty();
        actfocus = 0;
        }
}
#endif

#ifdef WINDOWK
// scroll the Window (must call show() first !)
// set nlines = 0 to clear
// set nlines < 0 to scroll down instead of up
void Window::scroll(Rectangle r,int color,int nlines)
{
    screenCoords(r);
    switch(visible(r))
        {
    case Error:
        ASSERT(!"Rectangle not inside window!");
    case Visible:
        setCurrentColorFromState(color);
        screen->scroll(r,currentColor,nlines);
        break;
    default:
        dirty(r);               // mark Rectangle as dirty
        break;
        }
}

int Window::visible(const Rectangle &r) const
{
    if(!this->encloses(r))
        return Error;                   // not inside window !

    Rectangle x(r);
    x &= *this;

    ObjectIterator next(*application);
    for(Window *w = (Window *)next(); w != this; w = (Window *)next())
        {
        ASSERT(w /* Window not in application */);

        if(w->intersects(x))
            return Hidden;              // partly or fully hidden
        }
    return Visible;                     // fully visible
}
#endif

#ifdef WINDOWL
void Window::drawbox(const Rectangle& r,const Rectangle &clip,int color,
        const BorderFlag s,const char *title)
{
    ASSERT(r.width() > 0 && r.height() > 0);

    int bordersize = 1;
    const char *boxch = boxChars[s & BFstyle];
    Rectangle at(r.topleft(),1,1);

    setCurrentColorFromState(color);
    if(r.width() == 1 || r.height() == 1)
        bordersize = 0;
    else
        {
        if(clip.encloses(at))
            fillrect(at,-1,boxch[Btopleft]);
        at.position(r.right(),r.top());
        if(clip.encloses(at))
            fillrect(at,-1,boxch[Btopright]);

        at.position(r.left(),r.bottom());
        if(clip.encloses(at))
            fillrect(at,-1,boxch[Bbotleft]);

        at.position(r.botright());
        if(clip.encloses(at))
            fillrect(at,-1,boxch[Bbotright]);
        }

    if(r.width() > bordersize + 1)
        {
        at.build(r.left()+bordersize,r.top(),r.right()-bordersize,r.top());

        if(clip.intersects(at))
            {
            at &= clip;
            fillrect(at,-1,boxch[Btop]);
            if((s & BFtitle) && *title)
                {
                Point p = r.topleft();
                int l = strlen(title);

                if(l > r.width() - 2)
                    l = r.width() - 2;
				switch(s & BFalignment)
					{
				case BFleft:
					p.x++;
					break;
				case BFright:
					p.x += r.width() - l - 1;
					break;
				default:
	                p.x += (r.width() - l) / 2;
					break;
					}
                if(l > 0)
                    printstring(p,at,-1,title,l);
                }
            }
        if(r.height() > 1)
            {
            at.build(r.left()+1,r.bottom(),r.right()-1,r.bottom());
            if(clip.intersects(at))
                {
                at &= clip;
                fillrect(at,-1,boxch[Bbot]);
                }
            }
        }

    if(r.height() > bordersize + 1)
        {
        at.build(r.left(),r.top()+1,r.left(),r.bottom()-1);

        if(clip.intersects(at))
            {
            at &= clip;
            fillrect(at,-1,boxch[Bleft]);
            }
        
        if(r.width() > 1)
            {
            at.build(r.right(),r.top()+1,r.right(),r.bottom()-1);
            if(clip.intersects(at))
                {
                at &= clip;
                fillrect(at,-1,boxch[Bright]);
                }
            }
        }
}
#endif

#ifdef WINDOWM
// Move text cursor to p
void Window::cursor(coord x,coord y)
{
    cursorpt.x = x;
    cursorpt.y = y;
    screenCoords(cursorpt);
    screen->cursor(cursorpt);
}
#endif

#ifdef WINDOWN
void Window::showcursor(int block)
{
    screen->showcursor(block);
}

void Window::hidecursor()
{
    screen->hidecursor();
}
#endif
